/*
	time.c

	Timers.

	Copyright (C) 2001 Sparkle Game Company
*/

#include "time.h"

/* 
	control registers 

	F E D C B A 9 8 7 6 5 4 3 2 1 0
					| |		  |	| |
					| |		  | \ /   frequency
					| |		  |  --------------
			enabled	| |		  | cascade
            --------  |		   --------
					  |
				IRQ	  |
			   -------
*/
#define TIMER0_CTRL		0x04000102
#define TIMER1_CTRL		0x04000106
#define TIMER2_CTRL		0x0400010A
#define TIMER3_CTRL		0x0400010E

#define TIMER_ENABLED	0x0080
#define TIMER_IRQ		0x0040
#define TIMER_CASCADE	0x0004


/*	timer frequencies

	frequency		speed
	---------		-----
	FREQ0:			16.78MHz
	FREQ1:			16.78MHz/64
	FREQ2:			16.78MHz/256
	FREQ3:			16.78MHz/1024
*/

#define TIMER_FREQ0		0x0000
#define TIMER_FREQ1		0x0001
#define TIMER_FREQ2		0x0010
#define TIMER_FREQ3		0x0011

/* global timers */
u16 *timer0 = (u16 *)0x04000100;
u16 *timer1 = (u16 *)0x04000104;
u16 *timer2 = (u16 *)0x04000108;
u16 *timer3 = (u16 *)0x0400010C;

void install_timers(void)
{
	*timer0 = 0;
	*timer1 = 0;
	*timer2 = 0;
	*timer3 = 0;

	/* setup timer speeds */
	*((u16 *)TIMER0_CTRL) = TIMER_ENABLED & TIMER_FREQ0;
	*((u16 *)TIMER1_CTRL) = TIMER_ENABLED & TIMER_FREQ1;
	*((u16 *)TIMER2_CTRL) = TIMER_ENABLED & TIMER_FREQ2;
	*((u16 *)TIMER3_CTRL) = TIMER_ENABLED & TIMER_FREQ3;
}