#include stdlib
#include "scriptdefs.h"

// *********************************
// ---------- AREA STATES ----------
// *********************************
#define NUM_AREAS		1
#define NUM_STATES		10

int area_states[NUM_AREAS][NUM_STATES];
int current_area = 0;

// ************************************************
// ---------- DIFFERENT AREAS (DEFINE'S) ----------
// ************************************************
#define AREA_TOWN		0
#define AREA_CASTLE1	1

// ****************************************
// ---------- PROCESSOR COMMANDS ----------
// ****************************************
#define COMMAND_INIT			99
#define COMMAND_SET_STATE		100
#define COMMAND_DIALOG			101

// *** area_processor ***
void area_processor(int area, int command, int extra1, int extra2);

// ************************************
// ---------- AREA FUNCTIONS ----------
// ************************************
void town_init();
void castle1_init();
void town_dialog(int, int);

int first_time = 1;
PLAYERC davin;

void town_init()
{
	NPC npc;
	ENEMY enemy1, enemy2;
		
	// only area 0 loads maps and plays music manually
	if(first_time)
	{
		map_load("town.fmp");
		music_play("forestgm.mid");
	}
	else
		music_play("forestgm.mid");

	npc = create_npc(32, 256, 16, 32, "char.bmp");
	
	if(area_states[AREA_TOWN][0] == 0)
		set_npc_dialog(npc, "You should visit the castle.");
	else
	if(area_states[AREA_TOWN][0] == 1)
		set_npc_dialog(npc, "So, how was the castle?");
	else
	if(area_states[AREA_TOWN][0] >= 2)
		set_npc_dialog(npc, "Stop visiting the castle!");

	enemy1 = create_enemy(160, 64, 16, 32, "char.bmp");
	enemy2 = create_enemy(208, 64, 16, 32, "char.bmp");

	if(first_time)
	{
		davin = create_party_member(112, 288, 16, 32, "char.bmp");
		first_time = 0;
	}
}

void castle1_init()
{
	NPC npc1, npc2, npc3;

	//load_music_events("castle_one.txt");

	area_states[AREA_TOWN][0]++;

	
	music_play("castlegm.mid");

	npc1 = create_npc(32, 176, 16, 32, "char.bmp");
	npc2 = create_npc(224, 176, 16, 32, "char.bmp");
	npc3 = create_npc(208, 48, 16, 32, "char.bmp");

}

void town_dialog(int x, int y)
{


}


int main()
{
	int i = 0;
	int j, k;

	for(j = 0; j < NUM_AREAS; j++)
		for(k = 0; k < NUM_STATES; k++)
			area_states[j][k] = 0;

	area_processor(0, COMMAND_INIT, 0, 0);
	
	while(!key_esc())
	{
		process_warp_events();
		process_music_events();
		read_joystick();
		
		if(up())
			move_party(UP, 2);
		
		if(down())
			move_party(DOWN, 2);
				
		if(left())
			move_party(LEFT, 2);
				
		if(right())
			move_party(RIGHT, 2);
	
		if(key_z())
			toggle_zoom();
	
		if(a())
		{
			i = talk();		/* try diplomacy first. :-) */
			if(!i)		
				attack();
		}

		map_draw();
		draw_characters();
		map_draw_fg();

		if(i)
		{
			display_npc_dialog(i);
			i = talk();
		}
				
		/* take a snap */
		if(key_s())
			screen_save();
		
		
		if(new_map != -1)
		{
			process_warp_events();
		//	process_music_events();

			area_processor(new_map, COMMAND_INIT, 0, 0);
			new_map = -1;
		}

		update_screen();
		limit_pc_cycles();
	}
}


// ==============
// area_processor
//
// This function processes commands for an area, sets states
// and executes area functions.
// ============================

void area_processor(int area, int command, int extra1, int extra2)
{
	switch(command)
	{
		case COMMAND_INIT:				// initialize the area
		{
			switch(area)
			{
				case AREA_TOWN:
				{
					town_init();
				} break;
				
				case AREA_CASTLE1:
				{
					castle1_init();
				} break;
			}
		} break;
		
		case COMMAND_SET_STATE:			// set an area state
		{
			area_states[area][extra1] = extra2;
		} break;

		case COMMAND_DIALOG:			// handle dialog
		{
			switch(area)
			{
				case 0:
				{
					town_dialog(extra1, extra2);
				} break;
			}
		} break;
	}
}