/*
    animation.c

    Last Coder : Greg

    Last Update : 6/16/01

    Last Changes : Added function header comments.

     2001 Sparkle

    Animation system.

*/

#include <allegro.h>
#include "animation.h"
#include "timer.h"

/* ANIMATION *rm_animation_create(BITMAP *bmp, int x_off, int y_off, int speed)

	Creates an animation starting at position (x_off, y_off) of bmp.  
	Speed can be a value 0-9, 0 being the slowest.

	Returns the animation, else NULL on error.
*/

ANIMATION *rm_animation_create(BITMAP *bmp, int x_off, int y_off, 
							int speed)
{
	int x, y, i;
	ANIMATION *a;

	if(speed < 0 || speed > 9)
		return NULL;
	
	if(x_off >= bmp->w || y_off >= bmp->h)
	{exit(0);
		return NULL;
	}
	
	/* verify the white border */
	if(getpixel(bmp, x_off, y_off) != makecol(255, 255, 255))
		return NULL;

	a = (ANIMATION *)malloc(sizeof(ANIMATION));
	
	if(!a)
		return NULL;
	
	/* start inside the white border */
	x = x_off + 1;
	y = y_off + 1;

	/* get width */
	while(getpixel(bmp, x, y) != makecol(255, 255, 255))
		x++;

	x--;
	a->width = x - x_off;

	/* get height */
	while(getpixel(bmp, x, y) != makecol(255, 255, 255))
		y++;

	y--;
	a->height = y - y_off;

	/* get number of frames */
	x = x_off;
	y = y_off;
	a->num_frames = 0;

	while(x < bmp->w && getpixel(bmp, x, y) == makecol(255, 255, 255))
	{
		x += a->width + 2;
		a->num_frames++;
	}

	
	/* grab the frames of animation */
	a->frames = (BITMAP **)malloc(sizeof(BITMAP *) * a->num_frames);
	
	if(!a->frames)
		return NULL;
	
	x = x_off + 1;
	y = y_off + 1;
	for(i = 0; i < a->num_frames; i++, x += a->width + 2)
	{
		a->frames[i] = create_bitmap(a->width, a->height);
		
		if(!a->frames[i])
			return NULL;

		blit(bmp, a->frames[i], x, y, 0, 0, a->width, a->height);
	}
	
	
	a->current_frame = 0;
	a->loop = 0;
	a->anim_speed = speed;
	a->anim_count = timer[speed];
	a->animating = 1;

	return a;
}

/*	int rm_animation_destroy(ANIMATION *a)

	Frees the memory that was used by the animation.
*/

int rm_animation_destroy(ANIMATION *a)
{
	int i;

	if(!a)
		return 0;

	for(i = 0; i < a->num_frames; i++)
	{
		destroy_bitmap(a->frames[i]);
	};

	free(a->frames);
	free(a);
	return 1;
}

/*	int rm_animation_update(ANIMATION *a)

	This function cycles through the frames in an animation.  Nothing will
	happen if the 'animating' variable is not set to a non-zero value.

	Returns 1 on success, else 0.
*/

int rm_animation_update(ANIMATION *a)
{
	if(!a->animating)
		return 0;
	
	/* has the timer updated? */
	if(a->anim_count != timer[a->anim_speed])
	{
		a->current_frame++;

		/* loop? */
		if(a->current_frame >= a->num_frames)
		{
			if(a->loop)
				a->current_frame = 0;
			else
				a->current_frame = a->num_frames - 1;
		}
	
		a->anim_count = timer[a->anim_speed];
	}

	return 1;
}