/*
    sprite.c

    Last Coder : Greg

    Last Update : 6/9/01

    Last Changes : Added rm_sprite_collision(SPRITE *a, SPRITE *b).

     2001 Sparkle

    Sprites.

*/

#include <allegro.h>
#include "animation.h"
#include "sprite.h"

/*	SPRITE *rm_sprite_create(char *gfx, int speed, int dist, int x, int y)

	Creates a sprite. It takes the following parameters:

	gfx		- the file to grab the frames from
	speed	- the speed the sprite animates at
	dist	- the sprite's moving distance
	x, y	- the sprite's location

	Returns NULL on error.
*/

SPRITE *rm_sprite_create(char *gfx, int speed, int dist, int x, int y)
{
	SPRITE *s;
	BITMAP *bmp;
	PALETTE pal;
	ANIMATION *tmp_a;
	int done = 0;
	int anim_count = 1;
	int pos_x, pos_y;

	s = (SPRITE *)malloc(sizeof(SPRITE));

	if(!s)
		return NULL;

	bmp = load_bitmap(gfx, pal);

	if(!bmp)
		return NULL;

	s->anim = NULL;
	s->motion_dist = dist;
	s->motion_speed = speed;
	s->x = x;
	s->y = y;
	s->num_sets = 0;

	pos_x = 0;
	pos_y = 0;

	/* grab the sets of animation */
	while(!done)
	{
		tmp_a = rm_animation_create(bmp, pos_x, pos_y, speed);
		
		if(!tmp_a)
		{
			done = 1;
			break;
		}
	
		/* add another set */
		s->anim = (ANIMATION **)realloc(s->anim, sizeof(ANIMATION *) * anim_count);
		
		if(!s->anim)
			return NULL;

		s->anim[anim_count - 1] = tmp_a;
		pos_y += s->anim[anim_count - 1]->height + 2;
		anim_count++;
	}

	destroy_bitmap(bmp);
	s->num_sets = anim_count - 1;
	s->current_set = 0;

	return s;
}

/*	int rm_sprite_destroy(SPRITE *s)

	Frees the memory used by a sprite.
	
	Returns 0 if the sprite is not valid.
*/

int rm_sprite_destroy(SPRITE *s)
{
	int i;

	if(!s)
		return 0;

	/* delete all the animation sets */
	for(i = 0; i < s->num_sets; i++)
	{
		rm_animation_destroy(s->anim[i]);
		s->anim[i] = NULL;
	}

	free(s->anim);
	s->anim = NULL;
	return 1;
}

/* int rm_sprite_draw(BITMAP *bmp, SPRITE *s)

	This function draws a sprite onto a bitmap.

	Returns 0 on failure.
*/

int rm_sprite_draw(BITMAP *bmp, SPRITE *s)
{
	int i;

	if(!bmp || !s)
		return 0;

	i = s->anim[s->current_set]->current_frame;
	
	draw_sprite(bmp, s->anim[s->current_set]->frames[i], s->x, s->y);
	return 1;
}

/* int rm_sprite_animate(SPRITE *s)

	Animates the sprite.  'animating' must be set to non-zero.

	Returns 0 on failure.
*/

int rm_sprite_animate(SPRITE *s)
{
	if(!s)
		return 0;
	
	if(!s->anim[s->current_set]->animating)
		return 0;

	rm_animation_update(s->anim[s->current_set]);
	
	return 1;
}

/* int rm_sprite_set_loop(SPRITE *s, int status)

	Sets or clears the looping flag of a sprite.  If status is 1,
	the sprite's animation will loop on the current set.  Zero 
	will clear the looping flag.
*/

int rm_sprite_set_loop(SPRITE *s, int status)
{
	if(!s)
		return 0;

	s->anim[s->current_set]->loop = status ? 1 : 0;
	return 1;
}

/* int rm_sprite_collision(SPRITE *a, SPRITE *b)

	Tests for collision between two sprites, a and b.

	Returns 1 if collision ocurred, else 0.
*/

int rm_sprite_collision(SPRITE *a, SPRITE *b)
{
	int dx, dy;
	int width, height;

	/* If sprite a is left of sprite b, use sprite a's width */
	
	if(a->x < b->x)
		width = a->anim[a->current_set]->width;
	else
		width = b->anim[b->current_set]->width;
	
	/* If sprite a is above of sprite b, use sprite a's height */
	
	if(a->y < b->y)
		height = a->anim[a->current_set]->height;
	else
		height = b->anim[b->current_set]->height;

	/* compute the overlap */
	dx = abs(a->x - b->x);
	dy = abs(a->y - b->y);

	/* collision found? */
	if(dx < width && dy < height)
		return 1;
	else 
		return 0;
}